"""Module with values of constants useful for solutions.

Note: dictionaries written in alphabetical order.

CONTENTS
--------
Fundamental constants:
    - k (float): Boltzmann's constant
    - Na (float): Avogadro's constant
    - e (float): elementary charge
    - C (float): speed of light in vacuum
Other constants:
    - R (float): ideal gas constant
    - Patm (float): atmospheric pressure in Pa
    - epsilon0: electric permittivity in vacuum
    - mu0: magnetic permeability in vacuum
Water properties:
    - Mw (float): molar mass of water in kg / mol
    - Tc (float): critical temperature in K
    - Pc (float): critical pressure in Pa
    - rhoc (float): critical density in kg / m^3
Solute properties as functions:
    - get_salt(formula): return Salt object (see Salt docstring and help)
    - molar_mass(formula). Input: formula (e.g. 'NaCl'), output M in kg / mol

SOURCES
-------
CRC Handbook of Physics and Chemistry:
    - Physical Constants of Inorganic Compounds
    https://hbcp.chemnetbase.com/faces/documents/04_02/04_02_0001.xhtml
    - Thermodynamic Properties of Aqueous Ions
    http://hbcponline.com/faces/documents/05_04/05_04_0001.xhtml
    - Recommended values of the fundamental physical constants
    http://hbcponline.com/faces/documents/01_01/01_01_0001.xhtml
    - Fixed-point properties of H20 and D20
    http://hbcponline.com/faces/documents/06_04/06_04_0001.xhtml
    - CODATA RECOMMENDED VALUES OF THE FUNDAMENTAL PHYSICAL CONSTANTS: 2018
    https://hbcp.chemnetbase.com/documents/18_01/18_01_0001.xhtml

IAPWS, Release on Surface Tension of Ordinary Water Substance
IAPWS, London, September 1994.

"""

# =========================== FUNDAMENTAL CONSTANTS ==========================

k = 1.380_649e-23   # Boltzmann's constant [J / K]
Na = 6.022_140_76e23    # Avogadro's constant [1/mol]
e = 1.602_176_634e-19   # elementary charge [C]
C = 299_792_458         # speed of light in vacuum [m/s]

# ============================= Other constants ==============================

R = k * Na
Patm = 101_325                   # atmospheric pressure [Pa]
epsilon0 = 8.854_187_8128e-12    # electric permittivity in vacuum
mu0 = 1.256_637_062_12e-6        # magnetic permeability in vacuum

# ============================== WATER PROPERTIES ============================

Mw = 18.015268e-3  # molar mass in kg / mol
Tc = 647.096  # critical temperature in K (IAPWS 2014)
Pc = 22.064e6  # critical pressure in Pa (CRC Handbook & IAPWS)
rhoc = 322    # critical density in kg/m^3 (CRC Handbook & IAPWS)

# =========================== SOLUTE/IONS PROPERTIES =========================


# Individual ions ------------------------------------------------------------


class Ion:
    """Representation of anions and cations"""

    def __init__(self, molecule, name, charge, molecular_weight):
        self.molecule = molecule  # e.g. 'Cl'
        self.name = name          # e.g. 'chloride'
        self.charge = charge      # e.g. -1
        self.molecular_weight = molecular_weight  # in Daltons

    @property
    def ion_type(self):
        return 'cation' if self.charge > 0 else 'anion'

    def __repr__(self):
        charge_str = ''
        if abs(self.charge) > 1:
            charge_str += f'{abs(self.charge)}'
        charge_str += ('+' if self.charge > 0 else '-')
        return f'{self.name.capitalize()} {self.ion_type} {self.molecule}[{charge_str}]'


cations = (
    Ion('Al', name='aluminium', charge=3, molecular_weight=26.982),
    Ion('Ca', name='calcium', charge=2, molecular_weight=40.078),
    Ion('K', name='potassium', charge=1, molecular_weight=39.098),
    Ion('Li', name='lithium', charge=1, molecular_weight=6.94),
    Ion('Mg', name='magnesium', charge=2, molecular_weight=24.305),
    Ion('Na', name='sodium', charge=1, molecular_weight=22.99),
)

anions = (
    Ion('Br', name='bromide', charge=-1, molecular_weight=79.904),
    Ion('Cl', name='chloride', charge=-1, molecular_weight=35.453),
    Ion('I', name='iodide', charge=-1, molecular_weight=126.904),
    Ion('NO3', name='nitrate', charge=-1, molecular_weight=62.005),
    Ion('SO3', name='sulfite', charge=-2, molecular_weight=80.063),
    Ion('SO4', name='sulfate', charge=-2, molecular_weight=96.063),
)

CATIONS = {cation.name: cation for cation in cations}
ANIONS = {anion.name: anion for anion in anions}


#  Solutes / salts -----------------------------------------------------------

class Solute():
    """Base class for solutes, including salts"""

    def __init__(self, formula, molar_mass):

        self.formula = formula
        self.molar_mass = molar_mass


class Salt(Solute):
    """Representation of salt solute and its characteristics"""

    def __init__(self, cation, anion):
        """Name is e.g. 'NaCl'"""
        self.cation = CATIONS[cation]
        self.anion = ANIONS[anion]
        self.ions = self.cation, self.anion
        self.charges = tuple((ion.charge for ion in self.ions))
        self.stoichiometry = tuple(abs(ion.charge) for ion in reversed(self.ions))
        self.molecular_weight = self._get_weight()      # in Daltons

        super().__init__(
            formula=self._get_formula(), 
            molar_mass=self.molecular_weight * 1e-3,    # in kg/mol
        )

    def _get_formula(self):
        formula = ''
        for ion, coeff in zip(self.ions, self.stoichiometry):
            formula += ion.molecule
            if coeff > 1:
                formula += str(coeff)
        return formula

    def _get_weight(self):
        """In Daltons"""
        ion_coeffs = zip(self.ions, self.stoichiometry)
        ws = [ion.molecular_weight * coeff for ion, coeff in ion_coeffs]
        return sum(ws)

    def __repr__(self):
        return f'{self.cation.name.capitalize()} {self.anion.name} {self.formula}'


salts = (
    Salt('aluminium', 'chloride'),
    Salt('calcium', 'chloride'),
    Salt('potassium', 'sulfate'),
    Salt('potassium', 'chloride'),
    Salt('potassium', 'iodide'),
    Salt('potassium', 'nitrate'),
    Salt('lithium', 'bromide'),
    Salt('lithium', 'chloride'),
    Salt('magnesium', 'chloride'),
    Salt('magnesium', 'sulfate'),
    Salt('sodium', 'sulfate'),
    Salt('sodium', 'chloride'),
    Salt('sodium', 'nitrate'),
)

SALTS = {salt.formula: salt for salt in salts}


class NeutralSolute(Solute):
    """Representation of neutral solute and its characteristics"""

    def __repr__(self) -> str:
        return f'Neutral solute: {self.formula.capitalize()}'

neutral_solutes = (
    NeutralSolute(formula='glycerol', molar_mass=92.0938e-3),
)

NEUTRAL_SOLUTES = {solute.formula: solute for solute in neutral_solutes}


# Useful constants and functions ---------------------------------------------


SOLUTES = {**SALTS, **NEUTRAL_SOLUTES}  # more solutes (possibly non-salts) can be added later


def get_solute(formula):
    """Get salt object from its formula"""
    try:
        return SOLUTES[formula]
    except KeyError:
        raise ValueError(f'{formula} not in available solutes: {list(SOLUTES)}')


def molar_mass(formula):
    """Return molar mass of solute compound in kg / mol."""
    solute = get_solute(formula=formula)
    return solute.molar_mass



"""Gathers the formulas for the density of KCl solutions.

see al_ghafri.py, krumgalz.py, tang.py moduled for more info
"""

from .al_ghafri import Density_KCl_AlGhafri_Base
from .krumgalz import Density_KCl_Krumgalz_Base
from .tang import Density_KCl_Tang_Base


class Density_KCl_AlGhafri(Density_KCl_AlGhafri_Base):
    """Already defined in Al Ghafri module"""
    default = True


class Density_KCl_Krumgalz(Density_KCl_Krumgalz_Base):
    """Already defined in Krumgalz module and not default here"""
    pass


class Density_KCl_Tang(Density_KCl_Tang_Base):
    """Already defined in Tang module and not default here"""
    pass


# ========================== WRAP-UP OF FORMULAS =============================

Density_KCl_Formulas = (
    Density_KCl_AlGhafri,
    Density_KCl_Krumgalz,
    Density_KCl_Tang,
)



"""Module for convenience functions related to solutions in pores"""


from aquasol.solutions import water_activity, aw_to_conc, convert


def pore_aw_from_drop_conc(
    drop_volume: float,
    pore_volume: float,
    solute: str = 'NaCl',
    T: float = 25.0,
    **concentration,
) -> float:
    """Calculate water activity in pores from a given drop volume and concentration

    Parameters
    ----------
    drop_volume : float
    pore_volume : float
        both drop_volume and pore_volume must be in the same units (e.g. µL)

    solute : str
        name of the solute, e.g. 'NaCl', 'glycerol', etc.

    T : float
        temperature in °C

    concentration : kwargs
        concentration input as required by aquasol, e.g.
        m=2 for molality of 2 mol / kg
        c=2300 for molarity of 2.3 mol / L
        x=0.12 for mole fraction of 12%
        etc.

    Returns
    -------
    float
        Water activity a_w in the pore solution (0 < a_w <= 1)
    """
    c_unit, = concentration.keys()
    c_value, = concentration.values()
    c_drop = convert(c_value, c_unit, 'c', T=T, solute=solute)
    c_pore = c_drop * drop_volume / pore_volume
    return water_activity(solute=solute, T=T, c=c_pore)


def drop_conc_from_pore_aw(
    a: float,
    drop_volume: float,
    pore_volume: float,
    out='m',
    solute: str = 'NaCl',
    T: float = 25.0,
) -> float:
    """Calculate concentration in droplet to achieve desired pore water activity

    Parameters
    ----------
    drop_volume : float
    pore_volume : float
        both drop_volume and pore_volume must be in the same units (e.g. µL)

    out : str
        Unit of output (e.g. 'x' for mole fraction, 'm' for molality, etc.)

    solute : str
        name of the solute, e.g. 'NaCl', 'glycerol', etc.

    T : float
        temperature in °C

    Returns
    -------
    float
        solute concentration in droplet
    """
    c_pore = aw_to_conc(a, solute=solute, T=T, out='c')
    c_drop = c_pore * pore_volume / drop_volume
    return convert(c_drop, 'c', out, solute=solute, T=T)



"""Gathers the formulas for the refractive index of KCl solutions.

See tang.py and tan.py modules for more info.
"""

from .tan import RefractiveIndex_KCl_Tan_Base
from .tang import RefractiveIndex_KCl_Tang_Base


class RefractiveIndex_KCl_Tan(RefractiveIndex_KCl_Tan_Base):
    """Already defined in tan module"""
    default = True


class RefractiveIndex_KCl_Tang(RefractiveIndex_KCl_Tang_Base):
    """Already defined in tang module and not default here"""
    pass


# ============================= WRAP-UP FORMULAS =============================

RefractiveIndexFormulas_KCl = (
    RefractiveIndex_KCl_Tan,
    RefractiveIndex_KCl_Tang,
)



"""Gathers the formulas for the refractive index of NaCl solutions.

See tang.py and tan.py modules for more info.
"""

from .tan import RefractiveIndex_NaCl_Tan_Base
from .tang import RefractiveIndex_NaCl_Tang_Base


class RefractiveIndex_NaCl_Tan(RefractiveIndex_NaCl_Tan_Base):
    """Already defined in tan module"""
    default = True


class RefractiveIndex_NaCl_Tang(RefractiveIndex_NaCl_Tang_Base):
    """Already defined in tang module and not default."""
    pass


# ============================= WRAP-UP FORMULAS =============================

RefractiveIndexFormulas_NaCl = (
    RefractiveIndex_NaCl_Tan,
    RefractiveIndex_NaCl_Tang,
)



"""Gathers the formulas for the refractive index of Na2SO4 solutions.

See tang.py module for more info.
"""

from .tang import RefractiveIndex_Na2SO4_Tang_Base


class RefractiveIndex_Na2SO4_Tang(RefractiveIndex_Na2SO4_Tang_Base):
    """Already defined in tang module"""
    default = True


# ============================= WRAP-UP FORMULAS =============================

RefractiveIndexFormulas_Na2SO4 = (
    RefractiveIndex_Na2SO4_Tang,
)



"""Refractive index of solutions according to Tang 1997

Source
------
Note:
    Tang 1997 cites other papers (e.g. Tang 1996) of his but compiles data
    in Table I. Some validity ranges (e.g. Na2SO4) are not correct in table I
    and I have taken them from the original sources.

Source
------
- Tang 1997 :
    Tang, I. N.
    Thermodynamic and optical properties of mixed-salt aerosols of atmospheric
    importance. Journal of Geophysical Research 102, 1883-1893 (1997).

Cites :

[NaCl]
Tang, I. N.
Chemical and size effects of hygroscopic aerosols on light scattering
coefficients.
Journal of Geophysical Research: Atmospheres 101, 19245-19250 (1996).

[Na2SO4]
Tang, I. N. & Munkelwitz, H. R.
Simultaneous Determination of Refractive Index and Density of an Evaporating
Aqueous Solution Droplet.
Aerosol Science and Technology 15, 201-207 (1991).

[Also Na2SO4]
Tang, I. N. & Munkelwitz, H. R.
Water activities, densities, and refractive indices of aqueous sulfates and
sodium nitrate droplets of atmospheric importance.
Journal of Geophysical Research 99, 18801 (1994).

[KCl]
Couldn't find the info in these other papers except Tang 1997
"""

from ....constants import Mw, molar_mass, SALTS
from ...general import SolutionFormula
from ..density.tang import Density_KCl_Tang_Base
from ..density.tang import Density_Na2SO4_Tang_Base
from ..density.tang import Density_NaCl_Tang_Base
from ..basic_conversions import basic_convert


IonicRefractions = {
    'Na': 0.86,
    'K': 3.21,
    'Cl': 8.09,
    'SO4': 13.44,
    'water': 3.717,
}


def ionic_refraction(solute):
    salt = SALTS[solute]
    R1 = IonicRefractions[salt.cation.molecule]
    R2 = IonicRefractions[salt.anion.molecule]
    nu1, nu2 = salt.stoichiometry
    return nu1 * R1 + nu2 * R2


density_classes = {
    'KCl': Density_KCl_Tang_Base,
    'Na2SO4': Density_Na2SO4_Tang_Base,
    'NaCl': Density_NaCl_Tang_Base,
}


class RefractiveIndex_Tang_Base(SolutionFormula):

    source = 'Tang'
    solute = None  # define in subclasses

    temperature_unit = 'C'
    temperature_range = (25, 25)

    concentration_unit = 'w'
    concentration_range = None  # define in subclasses

    with_water_reference = False

    def density(self, w, T):
        rho0, rho = density_classes[self.solute]().calculate(w=w, T=T)
        return rho

    def calculate(self, w, T):

        x = basic_convert(w, 'w', 'x', solute=self.solute)

        # ionic refraction
        Rw = IonicRefractions['water']
        Rs = ionic_refraction(self.solute)
        R = (1 - x) * Rw + x * Rs

        # molal volume
        rho = self.density(w=w, T=T)
        Ms = molar_mass(self.solute)
        V = ((1 - x) * Mw + x * Ms) / rho

        c = R / V * 1e-6  # factor 1e6 to go back to SI units

        return ((1 + 2 * c) / (1 - c)) ** (1 / 2)


# Note: I have adjusted the validity ranges below to the validity ranges
# of the density equation, because I did not see a stated validity range
# for the index of refraction


class RefractiveIndex_KCl_Tang_Base(RefractiveIndex_Tang_Base):
    solute = 'KCl'
    concentration_range = (0, 0.44)


class RefractiveIndex_Na2SO4_Tang_Base(RefractiveIndex_Tang_Base):
    solute = 'Na2SO4'
    concentration_range = (0, 0.68)


class RefractiveIndex_NaCl_Tang_Base(RefractiveIndex_Tang_Base):
    solute = 'NaCl'
    concentration_range = (0, 0.45)



"""
Gathers the formulas for the water activity of glycerol-water solutions.

Sources
-------

- Zhang, L., Grace, P. M. & Sun, D.-W. 
An accurate water activity model for glycerol solutions and its implementation on moisture sorption isotherm determination. 
Drying Technology 40, 2404–2413 (2022).
"""

from ...general import SolutionFormula

from .zhang import water_activity_glycerol


class WaterActivity_Glycerol_Zhang(SolutionFormula):

    source = 'Zhang'
    solute = 'glycerol'
    
    temperature_unit = 'K'
    temperature_range = (273.15, 373.15)

    concentration_unit = 'x'    
    concentration_range = (0, 97.5e-2)

    with_water_reference = False
    default = True

    def calculate(self, x, T):
        return water_activity_glycerol(x, T)
    

WaterActivityFormulas_Glycerol = (
    WaterActivity_Glycerol_Zhang,
)



""" Functions to calculate the density of glycerol-water solutions using Volk formulas.
Volk, A. & Kähler, C. J. Density model for aqueous glycerol solutions. Exp Fluids 59, 75 (2018).
"""

import numpy as np

def glycerol_density(T):
    """
    Calculate the density (kg/m^3) of pure glycerol based on temperature (K).
    """
    return 1273 - 0.612*T # kg/m^3

def water_density(T):
    """
    Calculate the density (kg/m^3) of pure water based on temperature (K).
    """
    return 1000 * (1 - np.abs((T - 3.98) / 615) ** 1.71) # kg/m^3

def density_water_glycerol(w, T):
    """
    Calculate the density (kg/m^3) of a glycerol-water solution based on mass fraction of glycerol and T.
    """
    A = 1.78e-6 * T**2 - 1.82e-4 * T + 1.41e-2
    kappa = 1 + A * np.sin(w**1.31 * np.pi)**0.81

    return kappa * (water_density(T) + w * (glycerol_density(T) - water_density(T)) / (w + (glycerol_density(T) / water_density(T)) * (1 - w))) # kg/m^3
